#ifdef SPH  
module cluster_class

  !
  ! Class definition and method for stellar cluster (multiple sources).
  !
  
  use sph_data_class, only: get_utime, get_time, get_umass, get_pt_mass, get_udist, get_rhon, &
       stellar_disc_exists
  use source_mod, only: SOURCETYPE
  use octal_mod, only: octal, subcellCentre, cellVolume, within_subcell
  use vector_mod
  implicit none

  public :: new,  &
       &    kill_all, &
       &    kill_only_stars, &
       &    get_nstar, &
       &    put_a_star, &
       &    get_a_star, &
       &    build_cluster, &
       &    assign_grid_values, &
       &    update_particle_list, &
       &    find_n_particle_in_subcell, &
       &    n_stars_in_octal, &
       &    remove_too_close_cells, &
       &    disc_density, &
       &    disc_intersects_subcell, &
       &    cell_too_close_to_star, &
       &    average_disc_density, &
       &    average_disc_density_fast, &
       &    max_disc_density_from_sample, &
       &    compute_scale_size, &
       &    include_disc, &
       &    assign_emission_bias, &
       &    restrict, &
       &    reassign_10K_temperature

  
  
  private :: int_cluster, int_cluster_alt,  compute_emission_bias
  
  
  type cluster
     private   ! It is better to be private! (The maintenance is easier later...)
     ! number of stars
     integer :: nstar
     ! stars
     type(sourcetype), pointer :: stars(:)  ! should be allocated with size=nstar
     !
     ! Flag to include accreation disc or not...
     logical :: disc_on  
  end type cluster
  
  ! overloading functions and creating the interface
  interface new
     module procedure int_cluster
     module procedure int_cluster_alt
  end interface
  

contains
  

  !==============================================================
  ! Constructor
  !==============================================================
  
  ! Initialize the object with default values.  

  subroutine int_cluster(this)
    implicit none
    type(cluster), intent(inout) :: this
    
    this%nstar= 1

    NULLIFY(this%stars)
    this% disc_on = .false.
    
  end subroutine int_cluster

  
  !
  ! Initializing with sph data
  !
  subroutine int_cluster_alt(this, disc_on)
    use sph_data_class, only: get_nptmass
    implicit none
    type(cluster), intent(inout) :: this
    logical, intent(in) :: disc_on
    ! Store important info from the data.    
    !
    this%nstar = get_nptmass()

    ! Allocate star array
    ALLOCATE(this%stars(this%nstar))

    !
    this%disc_on = disc_on

  end subroutine int_cluster_alt


  !==============================================================
  ! Destructor
  !==============================================================

  !
  !
  subroutine kill_all(this)
    implicit none
    type(cluster),intent(inout) :: this

    this%nstar=0
    DEALLOCATE(this%stars)
    NULLIFY(this%stars)

  end subroutine kill_all

  !
  ! Deallocates the stars array, but keeps other infomations
  subroutine kill_only_stars(this)
    implicit none
    type(cluster),intent(inout) :: this

    DEALLOCATE(this%stars)
    NULLIFY(this%stars)
    
  end subroutine kill_only_stars


  !==============================================================
  ! Accessors
  !==============================================================
  
  !
  ! Returns the number of stars
  !
  function get_nstar(this) RESULT(out)
    implicit none
    integer :: out
    type(cluster), intent(in) :: this

    out = this%nstar

  end function get_nstar


  !
  ! assigns a star to i-th position of stars array
  ! 
  subroutine put_a_star(this, a_star, i)
    implicit none
    type(cluster), intent(inout) :: this
    type(sourcetype), intent(in) :: a_star
    integer, intent(in) :: i
        
    this%stars(i) = a_star

  end subroutine put_a_star
    

  !
  ! Retrive a star stored in the i-th position of stars array.
  !
  function get_a_star(this, i) RESULT(out)
    implicit none
    type(sourcetype) :: out
    type(cluster), intent(in) :: this
    integer, intent(in) :: i

    ! Needs casting of a pointer???
    out = this%stars(i)

  end function get_a_star
    


  !==============================================================
  ! Tools 
  !==============================================================
  




  !
  ! Creates a cluster fully by assiging T, R, L and so on to each star
  ! 

  subroutine build_cluster(this, lambda_beg, lambda_end, iso_data, fix_source_R, fix_source_L, fix_source_T)
    use constants_mod, only: mSol, rSol, lSol, secsToYears
    USE parallel_mod, ONLY: torus_abort
    use isochrone_class
    use source_mod, only: fillspectrumBB, normalizedspectrum
    use sph_data_class, only: get_position_pt_mass
    implicit none
    type(cluster), intent(inout) :: this

    ! starting and the ending wavelength in Angstrome!
    real(double), intent(in) :: lambda_beg, lambda_end 
    !
    ! isochrone data to assign R, L, & T to stars
    type(isochrone), intent(in), optional :: iso_data  

    ! Use fixed values as an alternative to isochrone data.
    real(kind=8), optional, intent(in) :: fix_source_R ! Solar radii
    real(kind=8), optional, intent(in) :: fix_source_L ! L_sun
    real(kind=8), optional, intent(in) :: fix_source_T ! Kelvin
    
    !
    integer :: i, n
    type(sourcetype) :: a_star
    real(double) :: mass, age, radius, temperature
    real(double) :: luminosity
    real(double) :: length_units
    real(double) :: x, y, z
    
    n = get_nstar(this)

    ! using a function sph_data_class
    length_units = get_udist()/1.0d10 ! [10^10 cm]
    
    age = get_time()*get_utime() ! in [sec]
    ! converting to years
    age = age * secsToYears   ! [years]
    ! assigning stars
    do i = 1, n
       ! preparing the spectrum
       !

      if ( present(iso_data) ) then

         mass = get_pt_mass(i)*get_umass() ! [g]
         ! converting it to solar masses
         mass = mass/mSol ! [M_sun]

          ! finding the corresponding radius and the surface temperature of
          ! this star.
          ! -- using a routine in isochrone_class
          call  mass_age_to_temp_rad_lum(iso_data, mass, age, temperature, radius, luminosity)

       else

          ! Use fixed source properties
          if ( present(fix_source_R) ) then
             radius = fix_source_R * rSol * 1.0e-10 ! In torus length units
          else
             call torus_abort("build_cluster: neither fix_source_R nor iso_data is present")
          end if

          if ( present(fix_source_L ) ) then
             luminosity = fix_source_L * lSol
          else
             call torus_abort("build_cluster: neither fix_source_L nor iso_data is present")          
          end if

          if ( present(fix_source_T ) ) then
             temperature = fix_source_T
          else
             call torus_abort("build_cluster: neither fix_source_T nor iso_data is present")
          end if

       end if


       ! position of this star
       call get_position_pt_mass(i, x, y, z)
       x = x*length_units; y=y*length_units; z=z*length_units  ! [10^10cm]
       ! save it to a star
       a_star%luminosity = luminosity     ! erg/sec
!       a_star%luminosity = luminosity * 100.    ! erg/sec
!       write(*,*) "Luminosity fudged"
       a_star%radius = radius             ! 10^10cm
       a_star%teff = temperature          ! Kelvins
       a_star%position = Vector(x, y, z)  ! 10^10cm

       ! assiging a specturm to this star
       ! ( For now it is just BB specturm)
       ! using the routine in spectrum_mod.f90
       call fillSpectrumBB(a_star%spectrum, temperature, lambda_beg, lambda_end, 1000)
       call normalizedSpectrum(a_star%spectrum)
       
       ! Put this star in a array.
       ! Make sure that the arrays has been already allocated!
       this%stars(i) = a_star

    end do
       

  end subroutine build_cluster

  !
  ! Assigns various values which are needed for the run.
  ! This routine should be called only after the grid has been constructed, namely
  ! only in finishGrid routine in amr_mod.f90
  ! finishGrid makes a separate call to fillDensityCorners to fill density and 
  ! velocity corner values
  !
  subroutine assign_grid_values(thisOctal,subcell)

    use sph_data_class, only: Clusterparameter, sphData
    use constants_mod, only: mhydrogen
    use inputs_mod, only: tMinGlobal,doCOchemistry, x_D

    IMPLICIT NONE

    TYPE(octal), pointer :: thisOctal
    INTEGER :: subcell
    real(double), parameter :: density_crit = 1d13
    type(vector) :: point, clusterparam

    point = subcellcentre(thisOctal, subcell)
    ! Calculate density, temperature and H2 (if required) 
    clusterparam = Clusterparameter(point, thisoctal, subcell, theparam = 2)

    thisOctal%rho(subcell) = clusterparam%x
    if ( sphData%useSphTem ) then 
       thisOctal%temperature(subcell) =  max(real(clusterparam%y), tMinGlobal)
    else
       thisOctal%temperature(subcell) =  real(max(10.0_db, 10. * (thisOctal%rho(subcell) * density_crit)**(0.4)))
    end if

! Initialise to zero, will be set later in this subroutine if required
    if(associated(thisOctal%molabundance))thisOctal%molabundance = 0.0

! Set H2 number density from SPH particles if present
    if ( associated(thisOctal%nh2) ) then
       if (associated(sphData%rhoH2) ) then 
          thisOctal%NH2(subcell) = clusterparam%z / (2.0_db * mhydrogen)
       else
          thisOctal%nh2(subcell) = thisOctal%rho(subcell) / (2. * mhydrogen)
          if(doCOchemistry) then
             !       thisOctal%molAbundance(subcell) = 4e-10
             if(thisOctal%nh2(subcell) .lt. 3e4) thisOctal%molAbundance(subcell) = x_D ! drop fraction
             !       if(thisOctal%rho(subcell) .gt. 1.55885d-12) thisOctal%molAbundance(subcell) = 4e-9 !T > 30K
             if(thisOctal%temperature(subcell) .gt. 30.) thisOctal%molAbundance(subcell) = x_D ! drop fraction
          endif
       end if
    endif
    
    if (associated(thisOctal%microturb)) deallocate(thisoctal%microturb)
    
    if( associated(thisOctal%etaline)) thisOctal%etaLine(subcell) = 1.e-30
    if( associated(thisOctal%etaCont)) thisOctal%etaCont(subcell) = 1.e-30

! If sphData%rhoCO is in use then assume we want to use CO from SPH particles for abundance
    if ( associated (sphData%rhoCO) ) then 
       clusterparam = Clusterparameter(point, thisoctal, subcell, theparam = 3)
       thisOctal%molabundance(subcell) = real(max( ( clusterparam%x / ( 28.0_db * mhydrogen ) ) / thisOctal%nh2(subcell), 1d-37))
       write(115,*) thisOctal%molabundance(subcell), clusterparam%x, thisOctal%nh2(subcell)
    end if

  end subroutine assign_grid_values

  !
  ! Updates the sph particle linked list.
  !
  subroutine update_particle_list(thisOctal, subcell, newChildIndex)
    use sph_data_class, only: get_position_gas_particle
    implicit none
    TYPE(octal), INTENT(INOUT) :: thisOctal ! the parent octal 
!    type(octal), pointer :: thisOctal
    integer, intent(in) :: subcell       ! do loop index of thisOctal
    integer, intent(in) :: newChildIndex ! new child index
    !
    integer          :: np, i, j, k
    real(double) :: udist  ! units for distance
    real(double) :: x, y, z
    real(double) :: density_ave
    integer :: nparticle
    
    !
    ! Finding the number of particles in this subcell first
    call find_n_particle_in_subcell(nparticle, density_ave,&
         thisOctal, subcell)

    if (nParticle > 0) then
       ! Allocate the memory for new particle list of the child
       ALLOCATE(thisOctal%child(newChildIndex)%gas_particle_list(nparticle))
       
       ! # of particles in thisOctal
       np = SIZE(thisOctal%gas_particle_list)
       
       ! Units of length
       udist = get_udist() / 1.0d10  ! [10^10cm]
       
       k=0
!       !$OMP PARALLEL DEFAULT(NONE) &
!       !$OMP PRIVATE(i, j, x, y, z) &
!       !$OMP SHARED(thisOctal, subcell, k, udist, np, newChildindex)
!       !$OMP DO SCHEDULE(DYNAMIC,10)
       do i = 1, np
          ! extract the SPH index of particles in the parent cell.
          j = thisOctal%gas_particle_list(i)
          
          ! copy this value if this particle is in this child.
          ! Using a routine in sph_data_class. 
          call get_position_gas_particle(j, x, y, z)
          
          ! convert units
          x = x*udist; y = y*udist;  z = z*udist  ! [10^10cm]
          
          ! For 2D octals x is cylindrical polar r
          if (thisOctal%twoD) x = sqrt(x**2 + y**2)

          ! quick check to see if this gas particle
          ! belongs to this child cell.
          if ( within_subcell(thisOctal, subcell, x, y, z) ) then
!             !$OMP ATOMIC
             k = k+1
             ! add this particle index to this child using
             ! a routine in linked_list_class.f90
!             !$OMP ATOMIC
             thisOctal%child(newChildIndex)%gas_particle_list(k) = j
          end if
       end do
!       !$OMP END DO
!       !$OMP END PARALLEL

    endif
    
  end subroutine update_particle_list

    
  ! For a given octal object and sph_data_class object, this
  ! function returns the number of gas particles which belongs to a subcell this
  ! of this octal, and the average denisty ( in g/cm^3) of this cell.
  !  

  subroutine find_n_particle_in_subcell(n, rho_ave, node, subcell, rho_min, rho_max, v_min, v_max, n_pt)
    use sph_data_class, only: sphData, get_position_pt_mass, get_position_gas_particle, get_vel

    implicit none
    integer, intent(out) :: n                ! number of particles in the subcell
    integer, intent(out), optional :: n_pt   ! number of point masses in the subcell 
    real(double), intent(out) :: rho_ave ! average density of the subcell
    real(double), intent(out), optional :: rho_min, rho_max
    type(VECTOR), intent(out), optional :: v_min, v_max
    type(VECTOR) :: this_v
    real(double) :: vxmin, vxmax, vymin, vymax, vzmin, vzmax

    real(double) :: this_rho
    type(octal), intent(inout)    :: node
    integer, intent(in)           :: subcell ! index of the subcell

    integer :: npart
    integer :: i, j, counter
    real(double) :: x, y, z


    real(double), save :: umass, udist, udent  ! for units conversion  
    logical, save  :: first_time = .true.

    real(double), parameter    :: rho_null = 1.0e-30_db
    TYPE(VECTOR), parameter    :: v_null = VECTOR(-1.44d-8,-1.44d-8,-1.44d-8)
    TYPE(VECTOR), parameter    :: v_nullminus = VECTOR(1.44d-8,1.44d-8,1.44d-8)

    ! Carry out the initial calculations
    if (first_time) then       
       ! units of sphData
       umass = get_umass()  ! [g]
       udist = get_udist()  ! [cm]
       udent = umass/udist**3

       ! convert units
       udist = udist/1.0d10  ! [10^10cm]

       first_time = .false.
    end if

    if (associated(node%gas_particle_list)) then
       
       ! retriveing the number of the total gas particles
       ! in "node".
       ! -- using the function in linked_list_class.f90
       npart = SIZE(node%gas_particle_list)
       
       counter = 0
       rho_ave = 0.0d0
       if ( present(rho_min) ) rho_min = 1.0d30
       if ( present(rho_max) ) rho_max = 0.0d0
       if ( present(v_min) ) v_min = VECTOR(9.9d99,9.d99,9.d99)
       if ( present(v_max) ) v_max = VECTOR(-9.9d99,-9.9d99,-9.9d99)

!       !$OMP PARALLEL DEFAULT(NONE) &
!       !$OMP PRIVATE(i, j, x, y, z, this_rho, this_v) &
!       !$OMP SHARED(node, subcell, udist, npart, counter, rho_ave, rho_min, rho_max) &
!       !$OMP SHARED(v_min, v_max, vxmin, vymin, vzmin, vxmax, vymax, vzmax)
!       !$OMP DO SCHEDULE(DYNAMIC,10)
       do i=1, npart
          ! Retriving the sph data index for this particle
          j = node%gas_particle_list(i)
          
          ! retriving this posiiton of the gas particle.
          call get_position_gas_particle(j, x, y, z)
          
          ! convert units
          x = x*udist; y = y*udist; z = z*udist   ! [10^10cm]
          ! quick check to see if this gas particle is
          ! belongs to this cell.

          ! For 2D octals x is cylindrical polar r
          if (node%twoD) x = sqrt(x**2 + y**2)

          if ( within_subcell(node, subcell, x, y, z) ) then
           
!             !$OMP ATOMIC
             counter = counter + 1

             this_rho = get_rhon(j)
             if ( present(v_min) .or. present(v_max) ) then 
                this_v =  get_vel(j)
             end if

!             !$OMP ATOMIC
             rho_ave = rho_ave + this_rho

!             !$OMP CRITICAL
             if ( present(rho_min) ) rho_min = min(this_rho, rho_min)
             if ( present(rho_max) ) rho_max = max(this_rho, rho_max)
             if ( present(v_min) ) then
                vxmin = min(this_v%x, vxmin)
                vymin = min(this_v%y, vymin)
                vzmin = min(this_v%z, vzmin)
             endif
             if ( present(v_max) ) then
                vxmax = max(this_v%x, vxmax)
                vymax = max(this_v%y, vymax)
                vzmax = max(this_v%z, vzmax)
             endif
!             !$OMP END CRITICAL
          end if
          
       end do
!       !$OMP END DO
!       !$OMP END PARALLEL
       n = counter
    else
       n = 0
    endif
    
    ! Count the number of point masses if required
    if ( present(n_pt) ) then
       n_pt = 0
       do i=1, sphData%nptmass
          call get_position_pt_mass(i, x, y, z)
          ! convert units
          x = x*udist; y = y*udist; z = z*udist   ! [10^10cm]
          if (within_subcell(node, subcell, x, y, z)  ) n_pt = n_pt + 1
       end do
    end if

    if (n>0) then
       rho_ave = rho_ave/dble(n)

       rho_ave = rho_ave*udent  ! [g/cm^3]
       if ( present(rho_min) ) rho_min = rho_min * udent
       if ( present(rho_max) ) rho_max = rho_max * udent
       if ( present(v_min) ) v_min = VECTOR(vxmin,vymin,vzmin) * sphdata%codeVelocitytoTORUS
       if ( present(v_max) ) v_max = VECTOR(vxmax,vymax,vzmax) * sphdata%codeVelocitytoTORUS

    else
       rho_ave = rho_null
       if ( present(rho_min) ) rho_min = rho_null
       if ( present(rho_max) ) rho_max = rho_null
       if ( present(v_max) ) v_max = v_null
       if ( present(v_min) ) v_min = v_nullminus
    end if
    
  end subroutine find_n_particle_in_subcell


  function n_stars_in_octal(this, an_octal) result(out)
    use source_mod, only: source_within_octal
    implicit none
    integer :: out 
    type(cluster), intent(in) :: this 
    type(octal), intent(in) :: an_octal
    
    integer :: n, i 
    
    n = 0    
    do i = 1, this%nstar
       ! using a fucntion in source_mod
       if ( source_within_octal(this%stars(i), an_octal) ) n = n+1
    end do

    out = n
  end function n_stars_in_octal


  !
  ! Removes cells by setting the inflow flag to .false. and density to rho_small if the cells are  
  ! closer than a given radius (in 10^10cm) from star positions specified in cluster_class object.
  ! The value of axis allows the removed region to be either a sphere or a cylinder aligned with 
  ! one of the co-ordinate axes. 
  recursive subroutine remove_too_close_cells(thisCluster, thisOctal, R_max, removedMass, rho_small, axis)
    implicit none

    type(cluster), intent(in)   :: thisCluster    
    type(octal), pointer        :: thisOctal
    real(double), intent(in)    :: R_max   ! [10^10 cm]
    real(double), intent(inout) :: removedMass
    real(double), intent(in)    :: rho_small
    character(len=1), intent(in) :: axis
    !
    type(octal), pointer       :: pChild
    integer                    :: i           ! loop counter
    integer                    :: nc, subcell

    
    nc = thisOctal%nChildren    
    do subcell = 1, thisOctal%maxchildren
       if (thisOctal%hasChild(subcell)) then
          ! find the child
          do i = 1, nc
             if (thisOctal%indexChild(i) == subcell) then
                pChild => thisOctal%child(i)                
                call remove_too_close_cells(thisCluster, pChild, R_max, removedMass, rho_small, axis)
                exit
             end if
          end do
       else
          if ( thisOctal%cylindrical .and. (.not. thisOctal%splitAzimuthally) .and. subcell > 4) cycle
          if (.not. thisCluster%disc_on) then
             ! checks if it's too close 
             if (cell_too_close_to_star(thisCluster,thisOctal, subcell, R_max, axis) ) then
                thisOctal%inFlow(subcell) = .false.
                removedMass = removedMass + (thisOctal%rho(subcell)-rho_small) * cellVolume(thisOctal, subcell) * 1.0e30_db
                thisOctal%rho(subcell) = min(rho_small, thisOctal%rho(subcell))  ! set it to a very low density
             end if
          end if
       end if
    end do
    
    
  end subroutine remove_too_close_cells




  !
  ! This routine write the catalog of the stars in a cluster in 
  ! "catalog.dat".
  !
  subroutine write_catalog(this, filename)
    use messages_mod
    implicit none
    type(cluster),  intent(in) :: this
    character(len=*), optional :: filename
    
    integer, parameter :: LUOF=23
    integer :: i
    type(sourcetype) :: a_star
    real(double) :: mass
    real(double), parameter :: M_sun = 1.989e33         ! grams
    real(double), parameter :: L_sun = 3.85d33          !erg/s
    real(double), parameter :: R_sun = 6.96d0           ! 10^10cm

    ! 


    if ( present(filename) ) then
       call writeInfo("Stellar Cluster Catalog written in "//filename//"...",TRIVIAL)
       open(unit=LUOF, file=filename, status="replace")
    else
       call writeInfo("Stellar Cluster Catalog written in catalog.dat...",TRIVIAL)
       open(unit=LUOF, file="catalog.dat", status="replace")
    end if

    ! writing the header
    write(LUOF,'(a)')       "#  This file was written by cluster::write_catalog."
    write(LUOF,'(a)')       "#" 
    write(LUOF,'(a,1x,i15)') "# nstar =", this%nstar
    write(LUOF,'(a,1x,1PE12.3,1x,a10)') &
         &                  "# age   =", get_time()*get_utime()/3.1536000d7, " years old"
    write(LUOF,'(a)')       "#" 
    write(LUOF,'(a)')       "#" 
    write(LUOF,11)          "# ", "index", "x[10^cm]", "y[10^cm]", "z[10^cm]", &
         "mass [Msun]", "L [Lsun]", "radius  [Rsun]", "temperature [K]"
11  format(a,1x,a5, 7(2x,a15))
12  format(2x,i5, 7(1x,1PE16.7))
    write(LUOF, '(a)')      "#-----------------------------------------------"//&
         &"---------------------------------------------------------------------------------"
    do i = 1, this%nstar
       a_star = get_a_star(this,i)
       mass = get_pt_mass(i)*get_umass() ! [g]
       write(LUOF,12) i, a_star%position%x, a_star%position%y, a_star%position%z, &
            mass/M_sun, a_star%luminosity/L_sun, a_star%radius/R_sun, a_star%teff
    end do

    close(LUOF)

  end subroutine write_catalog



  ! Given a position (xpos, ypos and zpos) in 10^10 cm, this routine
  ! will return the density [g/cm^3] and the length scale (sizescale).
  !
  ! Note: This routine has been slightly changed from the original 
  !       one (discdensity.f) created by Matthew Bates. 
  !
  ! Matthew's comments:
  !     Adds discs around sink particles
  !     Disc is standard solution Sigma=R^-3/4, H=R^9/8 with H/R=0.05 at Rmax
  !     Returns density at a point in parameter space (0 if not within a disc)
  !     and gives a characteristic sizescale for grid construction.
  
  function disc_density(xpos,ypos,zpos, young_cluster ,sizescale) RESULT(density_out)
    use sph_data_class, only: get_stellar_disc_parameters
    implicit none
    real(double) :: density_out  ! in [g/cm^3]
    real(double), intent(in) :: xpos,ypos,zpos  ! should be in [10^10cm]
    type(cluster), intent(in)    :: young_cluster
    real(double), intent(out):: sizescale       ! should be in [10^10cm]
    !
    logical,save :: first_time = .true.
    real(double), parameter :: pi = 3.14159265d0
    integer :: nstar, i 
    real(double) :: holesize
    real(double) :: discrad, discmass, unitx, unity, unitz
    real(double) :: spinx, spiny, spinz, starrad, x, y, z
    real(double) :: xposi, yposi, zposi, rad, radius2, radius
    real(double) :: unorm, zdist, height, densitymidplane
    type(sourcetype) :: a_star
    ! maximum density allowed
!    real(double), parameter :: rho_max = 1.0d-12   ! g/cm^3
    real(double), parameter :: rho_max = 2.5d50   ! g/cm^3 (equivalent to no threshold)
    !
    real(double) :: s2 !, s1  
    logical :: logscale
!    real(double) :: Rgap1 = 1000.0 ! 10^10 cm
!    real(double) :: Rgap2 = 2000.0 ! 10^10 cm
    

    ! quick check
    if (first_time) then
       if (stellar_disc_exists()) then ! OK
          first_time = .false.
       else
          write(*,*) "Error:: You did not read in the stellar disc data! &
               &[cluster::disc_density]."
          stop
       end if
    end if
    
    !
    !  Inner disc radius is 3 times stellar radius (starrad(i) in unit of R_sun)
    !  required to be in units of 10^10 cm (i.e. times 7x10^10)
    !
    nstar = get_nstar(young_cluster)
!    holesize = 3.0d0 * 7.0d+00  !this is the original choice.
    holesize = 6.0d0 * 7.0d+00  !this is the original choice.
!    holesize = 12.0d0 * 7.0d00   ! 12 times the radius of a star



    !
    density_out = 1.0d-27
    sizescale = 1.0d+10
    DO i = 1, nstar   
       a_star = get_a_star(young_cluster,i)
       
       ! retrive some disc parameters from sphData
       call get_stellar_disc_parameters(i, discrad, discmass, &
            spinx, spiny, spinz)

       ! radius of this star
       starrad = a_star%radius  ! [10^10cm]
       starrad = starrad/7.0d0  ! [R_sun]
       x = a_star%position%x
       y = a_star%position%y
       z = a_star%position%z
       
       xposi = xpos-x
       yposi = ypos-y
       zposi = zpos-z
       radius2 = xposi**2 + yposi**2 + zposi**2

       radius = SQRT(radius2)
       IF (radius.LT.discrad) THEN          
          IF (radius.LT.1.1d0*(holesize*starrad)) THEN
!          IF (radius.LT.1.1d0*(holesize*starrad) .or. &
!               ( radius < Rgap2 .and. radius >Rgap1 ) ) THEN
             sizescale = holesize*starrad
             IF (radius > (holesize*starrad/2.0d0)) sizescale=sizescale/2.0d0
          ELSE
             rad = radius/discrad
             !  Distance above disc plane is obtained by doing a dot product with the
             !     unit vector giving the disc orientation
             unitx = spinx
             unity = spiny
             unitz = spinz
             unorm = SQRT(unitx**2 + unity**2 + unitz**2)
             zdist = (xposi*unitx + yposi*unity + zposi*unitz)/unorm
             !  Set density and characteristic size scale at position
             height = 0.05d0*discrad*rad**(1.125d0)

             densitymidplane = 5.0d0*discmass/(8.0d0*pi*discrad**2)*  &
                  rad**(-0.75d0)
             density_out = densitymidplane*   &
                  EXP(-zdist**2/(2.0d0*height**2))/(SQRT(2.0d0*pi)*height)
             if (density_out>1.0d-40) then
!                ! original
!                sizescale=0.1d0*height*(densitymidplane/density_out)**(1.0d0/3.0d0)
!                sizescale=1.0d-2*sizescale

!                ! finding the scale size at this location using 
!                ! the function in this module.

!                s1 = compute_scale_size(radius, holesize*starrad, discrad, &
!                     5.0d0*a_star%radius, 1.0d6, logscale)
!                s2 = s1




!!
!! Produces the 90 Mb grid around star 4
!!
                logscale = .false.
                s2 = compute_scale_size(ABS(zdist),0.0d0,  20.0d0*height, &
                    1.0d0*height,  5.0d0*height, logscale)
                sizescale = s2

!                logscale = .false.
!                s2 = compute_scale_size(ABS(zdist),0.0d0,  20.0d0*height, &
!                    1.0d0*height,  4.0d0*height, logscale)
!                sizescale = s2


! 
! This scaling creates 330Mb grid around star 4 and 660M around  star 46
!  
!                logscale = .false.
!                s2 = compute_scale_size(ABS(zdist),0.0d0,  8.0d0*height, &
!                    2.0d0*height,  1.0d0*height, logscale)
!                sizescale = s2


!!
!! For Tim Naylor
!!                logscale = .false.
!                s2 = compute_scale_size(ABS(zdist),0.0d0,  16.0d0*height, &
!                    4.0d0*height,  2.0d0*height, logscale)
!                sizescale = s2

!                logscale = .false.
!                s2 = compute_scale_size(ABS(zdist),0.0d0,  2.0d0*height, &
!                    1.0d0*height,  0.5d0*height, logscale)
!                sizescale = s2




!
!                ! limit the smallest and largest sizes
                sizescale = MAX(sizescale, a_star%radius*1.0d0)
                sizescale = MIN(sizescale, 2.0d3)                

                 ! else takes the default density from the top.
             end if
          ENDIF
          !  Given position can only be inside 1 disc at most (discs cannot overlap) 
          GOTO 10
       ENDIF
    END DO
10  CONTINUE    

    ! converting units from g/(10^10cm)^3 to g/cm^3

    density_out = density_out/1.0d30   ! [g/cm^3]

    ! assuming dust density is 100 times smaller than the gas density
    density_out = density_out/1.0d2   ! [g/cm^3]


    ! cutoff density
    ! maximum density allowed
    if (density_out>rho_max) density_out = rho_max ! [g/cm^3]
    

!    !
!    ! Emulating the larger grain size....
    !   
!    density_out = density_out/1.0d3   ! [g/cm^3]
!
!    ! limiting the minimum density
!    ! should be removed later
!    if (density_out<1.0d-20) density_out = 1.0d-20   ! [g/cm^3]


  end function disc_density




  !
  ! For a given octal, this routine checks if any of the stellar disk 
  ! in a cluster crosses a subcell of a octal.
  ! As a quick check, we consider the disk as a sphere (with r=Rdisc) 
  ! and the cell as a sphere with r=SQRT(3)*(d/2) where d is the subcell size
  function disc_intersects_subcell(this, octal_in, subcell) RESULT(out)
    use sph_data_class, only: get_stellar_disc_parameters
    implicit none
    logical :: out 
    
    type(cluster), intent(in) :: this
    type(octal), intent(in) :: octal_in
    integer, intent(in) :: subcell

    real(double) :: x, y, z, xc, yc, zc
    type(sourcetype) :: a_star
    integer :: i, nstar
    real(double) :: dummy_d, R_disc, R2_disc, R2_cell, d2
    TYPE(vector)     :: cellCenter

    nstar = get_nstar(this)
    
    R2_cell = 3.0d0*octal_in%subcellSize*octal_in%subcellSize ! square of the cell radius    

    ! position if the cell
    cellCenter = subcellCentre(octal_in,subcell)
    xc=dble(cellCenter%x); yc=dble(cellCenter%y); zc=dble(cellCenter%z)
   
    out = .false.
    do i = 1, nstar
       a_star = get_a_star(this,i)
       ! retrive some disc parameters from sphData
       call get_stellar_disc_parameters(i, R_disc, dummy_d, &
            dummy_d, dummy_d, dummy_d)       
       
       ! position of the star
       x = a_star%position%x
       y = a_star%position%y
       z = a_star%position%z
       
       ! The square of the distance between the star and the cell center.
       d2 = (x-xc)**2 + (y-yc)**2 + (z-zc)**2
       
       R2_disc = R_disc**2

       if (d2 < R2_disc+R2_cell) then 
          ! the two sphere intersect
          out = .true.
          exit  ! exit this loop.
       end if          
    end do
     
  end function disc_intersects_subcell




  !
  ! For a given octal, this routine checks if the center of the cell is with
  ! in a given radius from stars.
  ! The removed reqion can be either a sphere or a cylinder, depending on the axis argument.
  function cell_too_close_to_star(this,octal_in, subcell, R_max, axis) RESULT(out)
    implicit none
    logical :: out 
    
    type(cluster), intent(in) :: this
    type(octal), intent(in) :: octal_in
    integer, intent(in) :: subcell
    real(double), intent(in) :: R_max    ! limiting distance in [10^10cm]
    character(len=1), intent(in) :: axis ! Type of region (sphere or cylinder)

    real(double) :: x, y, z, xc, yc, zc
    type(sourcetype) :: a_star
    integer :: i, nstar
    real(double) :: R_sq, R_max_sq
    TYPE(vector)     :: cellCenter

    nstar = get_nstar(this)

    ! position if the cell
    cellCenter = subcellCentre(octal_in,subcell)
    xc=dble(cellCenter%x); yc=dble(cellCenter%y); zc=dble(cellCenter%z)

    R_max_sq = R_max**2

    out = .false.
    do i = 1, nstar
       a_star = get_a_star(this,i)
       
       ! position of the star
       x = a_star%position%x
       y = a_star%position%y
       z = a_star%position%z
       
       ! The square of the distance between the star and the cell center
       ! Use either a spherical or a cylindrical distance as required. 
       select case (axis)
       case ("s")
          R_sq = (x-xc)**2 + (y-yc)**2 + (z-zc)**2
       case ("x")
          R_sq = (y-yc)**2 + (z-zc)**2
       case ("y")
          R_sq = (x-xc)**2 + (z-zc)**2
       case ("z")
          R_sq = (x-xc)**2 + (y-yc)**2
       end select
       
       if (R_sq < R_max_sq) then 
          ! the two regions intersect
          out = .true.
          exit  ! exit this loop.
       end if          
    end do
     
  end function cell_too_close_to_star





  !
  ! Find the average density of the subcell by randomly picking points 
  ! in the subcell. Density is eveualated according to "disc_density" function 
  ! in this module, but it does not include the density of from the sph particles.
  !
  function average_disc_density(node, subcell, stellar_cluster, &
       scale_length) RESULT(out)
    implicit none
    real(double) :: out     ! out put density in [g/cm^3]
    type(octal), intent(inout)    :: node
    integer, intent(in)           :: subcell ! index of the subcell
    type(cluster), intent(in)     :: stellar_cluster
    real(double), optional, intent(inout) :: scale_length  ! [10^10cm]

    !
    integer, parameter :: nsample =300
    real(double) :: d, x, y, z, xc, yc, zc, rho_disc_ave, r, dummy_d, rho_sample
    type(vector)     :: cellCenter
    logical, save :: first_time = .true.
    integer :: i, n
    
    ! quick check for the first time
    if (first_time) then
       if (stellar_disc_exists()) then ! OK
          first_time = .false.
       else
          write(*,*) "Error:: You did not read in the stellar disc data! &
               &[cluster::average_disc_density]."
          stop
       end if
    end if
    
    !
    rho_disc_ave =1.0d-30
    if (stellar_disc_exists()) then
       d = node%subcellsize/2.0d0
       cellCenter = subcellCentre(node, subcell)
       xc=dble(cellCenter%x); yc=dble(cellCenter%y); zc=dble(cellCenter%z)
       
       n = 0       
       do i = 1, nsample
          ! picking a random position in the subcell
          call randomNumberGenerator(getDouble=r)
          x =  xc - d + r*2.0d0*d
          call randomNumberGenerator(getDouble=r)
          y =  yc - d + r*2.0d0*d
          call randomNumberGenerator(getDouble=r)
          z =  zc - d + r*2.0d0*d               
          ! evaluate the disc density density
          ! Taking the max value samples
          rho_sample = disc_density(x, y, z, stellar_cluster, dummy_d)
          if (rho_sample > rho_disc_ave) rho_disc_ave  = rho_sample

       end do
       


    end if
    
    ! Finding the scale_length
    if (present(scale_length)) then
       dummy_d = disc_density(xc, yc, zc, stellar_cluster, scale_length)
    end if


    out = rho_disc_ave  ! this should be in [g/cm^3] as in disc_density.

  end function average_disc_density




  !
  ! Find the average density of the subcell using the 8 corners points and 
  ! the center point of the subcell.
  ! Density is eveualated according to "disc_density" function 
  ! in this module, but it does not include the density of from the sph particles.
  !
  function average_disc_density_fast(node, subcell, stellar_cluster, &
       scale_length) RESULT(out)
    implicit none
    real(double) :: out     ! out put density in [g/cm^3]
    type(octal), intent(inout)    :: node
    integer, intent(in)           :: subcell ! index of the subcell
    type(cluster), intent(in)     :: stellar_cluster
    real(double),  intent(inout) :: scale_length

    !
    real(double) :: d, d2, xc, yc, zc, rho_disc_ave, dummy_d
    type(vector)     :: cellCenter
    logical, save :: first_time = .true.
    real(double) :: c0, c1, c2, c3, c4, c5, c6, c7, c8
    real(double) :: c9, c10, c11, c12, c13, c14, c15, c16
    real(double) :: c17, c18, c19, c20, c21, c22
    
    ! quick check for the first time
    if (first_time) then
       if (stellar_disc_exists()) then ! OK
          first_time = .false.
       else
          write(*,*) "Error:: You did not read in the stellar disc data! &
               &[cluster::daverage_disc_density]."
          stop
       end if
    end if
    
    !
    rho_disc_ave =0.0d0
    if (stellar_disc_exists()) then
       d = node%subcellsize/3.0d0
       d2 = d*2.0d0
       cellCenter = subcellCentre(node, subcell)
       xc=dble(cellCenter%x); yc=dble(cellCenter%y); zc=dble(cellCenter%z)
       
       c0 =  disc_density(xc, yc, zc, stellar_cluster, scale_length)

       c1 =  disc_density(xc+d, yc+d, zc+d, stellar_cluster, dummy_d)
       c2 =  disc_density(xc+d, yc+d, zc-d, stellar_cluster, dummy_d)
       c3 =  disc_density(xc+d, yc-d, zc-d, stellar_cluster, dummy_d)
       c4 =  disc_density(xc+d, yc-d, zc+d, stellar_cluster, dummy_d)
       c5 =  disc_density(xc-d, yc+d, zc+d, stellar_cluster, dummy_d)
       c6 =  disc_density(xc-d, yc+d, zc-d, stellar_cluster, dummy_d)
       c7 =  disc_density(xc-d, yc-d, zc-d, stellar_cluster, dummy_d)
       c8 =  disc_density(xc-d, yc-d, zc+d, stellar_cluster, dummy_d)

       c9  =  disc_density(xc+d2, yc+d2, zc+d2, stellar_cluster, dummy_d)
       c10 =  disc_density(xc+d2, yc+d2, zc-d2, stellar_cluster, dummy_d)
       c11 =  disc_density(xc+d2, yc-d2, zc-d2, stellar_cluster, dummy_d)
       c12 =  disc_density(xc+d2, yc-d2, zc+d2, stellar_cluster, dummy_d)
       c13 =  disc_density(xc-d2, yc+d2, zc+d2, stellar_cluster, dummy_d)
       c14 =  disc_density(xc-d2, yc+d2, zc-d2, stellar_cluster, dummy_d)
       c15 =  disc_density(xc-d2, yc-d2, zc-d2, stellar_cluster, dummy_d)
       c16 =  disc_density(xc-d2, yc-d2, zc+d2, stellar_cluster, dummy_d)

       c17 =  disc_density(xc+d2, yc, zc, stellar_cluster, dummy_d)
       c18 =  disc_density(xc-d2, yc, zc, stellar_cluster, dummy_d)
       c19 =  disc_density(xc, yc+d2, zc, stellar_cluster, dummy_d)
       c20 =  disc_density(xc, yc-d2, zc, stellar_cluster, dummy_d)
       c21 =  disc_density(xc, yc, zc+d2, stellar_cluster, dummy_d)
       c22 =  disc_density(xc, yc, zc-d2, stellar_cluster, dummy_d)

       
       rho_disc_ave = rho_disc_ave + max(c0, c1, c2, c3, c4, c5, c6, c7, c8, &
            c9, c10, c11, c12, c13, c14, c15, c16, c17, c18, c19, c20, c21, c22)

!       rho_disc_ave = rho_disc_ave + (c0+c1+c2+c3+c4+c5+c6+c7+c8 &
!            +c9+c10+c11+c12+c13+c14+c15+c16+c17+c18+c19+c20+c21+c22)/23.0d0
       
    end if

    out = rho_disc_ave  ! this should be in [g/cm^3] as in disc_density.

  end function average_disc_density_fast



  !
  ! Find the average density of the subcell by randomly picking points 
  ! in the subcell. Density is eveualated according to "disc_density" function 
  ! in this module, but it does not include the density of from the sph particles.
  !
  function max_disc_density_from_sample(node, subcell, stellar_cluster, &
       scale_length) RESULT(out)
    implicit none
    real(double) :: out     ! out put density in [g/cm^3]
    type(octal), intent(inout)    :: node
    integer, intent(in)           :: subcell ! index of the subcell
    type(cluster), intent(in)     :: stellar_cluster
    real(double), optional, intent(inout) :: scale_length  ! [10^10cm]

    !
    integer, parameter :: nsample =300
    real(double) :: d, x, y, z, xc, yc, zc, rho_disc_max, r, dummy_d
    real(double) :: rho_tmp
    type(vector)     :: cellCenter
    logical, save :: first_time = .true.
    integer :: i
    
    ! quick check for the first time
    if (first_time) then
       if (stellar_disc_exists()) then ! OK
          first_time = .false.
       else
          write(*,*) "Error:: You did not read in the stellar disc data! &
               &[cluster::average_disc_density]."
          stop
       end if
    end if
    
    !
    rho_disc_max =0.0d0
    if (stellar_disc_exists()) then
       d = node%subcellsize/2.0d0
       cellCenter = subcellCentre(node, subcell)
       xc=dble(cellCenter%x); yc=dble(cellCenter%y); zc=dble(cellCenter%z)
       
       do i = 1, nsample
          ! picking a random position in the subcell
          call randomNumberGenerator(getDouble=r)
          x =  xc - d + r*2.0d0*d
          call randomNumberGenerator(getDouble=r)
          y =  yc - d + r*2.0d0*d
          call randomNumberGenerator(getDouble=r)
          z =  zc - d + r*2.0d0*d
          ! evaluate the disc density density
          rho_tmp = disc_density(x, y, z, stellar_cluster, dummy_d)
          rho_disc_max = MAX(rho_disc_max, rho_tmp)
       end do
       
    end if
    
    ! Finding the scale_length
    if (present(scale_length)) then
       dummy_d = disc_density(xc, yc, zc, stellar_cluster, scale_length)
    end if


    out = rho_disc_max  ! this should be in [g/cm^3] as in disc_density.

  end function max_disc_density_from_sample


  !
  ! Find the scale size of a disc for given sizes of 
  ! minimum and maximum size scales allowed (ds_min and ds_max), 
  ! and minumum and maximum distance (e.g. radius or height above the disk..), i.e.
  ! s_min and s_max.
  function compute_scale_size(s, s_min, s_max, ds_min, ds_max, logscale) RESULT(out)
    implicit none
    real(double) :: out
    ! distance from center of a star or from the disk plane
    real(double), intent(in) :: s,  s_min, s_max, ds_min, ds_max
    logical, intent(in) :: logscale
    real(double)  :: p,ss

    ss=s
    if (logscale) then
       if (s<s_min) ss =s_min
       if (s>s_max) ss =s_max    
       p = LOG(ds_max/ds_min) / (s_max - s_min)
       out = ds_min * EXP(p*(ss-s_min))
    else ! linear scale
       if (s<s_min) ss =s_min
       if (s>s_max) ss =s_max   
       p = (ds_max-ds_min) / (s_max - s_min)
       out = ds_min +  p*(ss-s_min)       
    end if

  end function compute_scale_size


  ! 
  ! Just returns if this cluster includes accreation disc or not
  function include_disc(this) RESULT(out)
    implicit none 
    logical :: out
    type(cluster), intent(in) :: this ! cluster

    out = this%disc_on
  end function include_disc

       



  !
  ! Given a octal and subcell index and  a grid, and a wavelength in 
  ! the local frame, this routine computes
  ! the emission bias for contiuum and line photons.
  !
  subroutine compute_emission_bias(thisOctal,subcell, grid, lambda, lambda_array, n_lambda)
    use gridtype_mod, only: GRIDTYPE
    use utils_mod, only: hunt
    IMPLICIT NONE
    
    TYPE(octal), intent(inout) :: thisOctal
    INTEGER, INTENT(IN) :: subcell
    TYPE(gridtype), INTENT(INOUT) :: grid
    real, intent(in) :: lambda ! local wavelength in Angstrom
    integer, intent(in) :: n_lambda
    real, intent(in) :: lambda_array(n_lambda)

    !
    real(double) :: d    ! size of the octal
    real(double) :: tau  ! optical depth across the cell
    integer :: ilambda
    real(double) :: kappaScaReal, KappaAbsReal


    d = thisOctal%subcellSize

    call hunt(lambda_array, n_lambda, lambda, ilambda)

    ! abs and scattering coeff at this wavelength
    if (.not.grid%oneKappa) then
       kappaScaReal = thisOctal%kappaSca(subcell,iLambda)
       kappaAbsReal = thisOctal%kappaAbs(subcell,iLambda)
    else
       kappaScaReal = grid%oneKappaSca(thisOctal%dustType(subcell),iLambda)*thisOctal%rho(subcell)
       kappaAbsReal = grid%oneKappaAbs(thisOctal%dustType(subcell),iLambda)*thisOctal%rho(subcell)
    endif


    if (thisOctal%inFlow(subcell)) then
       tau = d * dble(kappaAbsReal + kappaScaReal)
       if (tau>100.0d0)  tau=100.d0
       if (tau<0.001d0)  tau=0.001d0
    else
       tau = 1.0d100
    end if 

    if (tau <= 0.0d0) then
       write(*,*) "Error:: tau <=0 in [cluster_class::compute_emission_bias]."
       stop
    end if
    
    thisOctal%biasCont3D(subcell) = 1.0d0/tau
    thisOctal%biasLine3D(subcell) = 1.0d0  ! for now this is just 1
       
  end subroutine compute_emission_bias


  !
  ! Recursively assigns the emission bias to all octals.
  !
  recursive subroutine assign_emission_bias(thisOctal,grid, lambda, lambda_array, n_lambda)
    use gridtype_mod, only: GRIDTYPE
    IMPLICIT NONE
    
    TYPE(octal), pointer :: thisOctal
    TYPE(gridtype), INTENT(INOUT) :: grid
    real, intent(in) :: lambda ! local wavelength in Angstrom
    integer, intent(in) :: n_lambda
    real, intent(in) :: lambda_array(n_lambda)
    
    type(octal), pointer   :: child
    integer :: subcell, ichild
     
    do subcell = 1, 8, 1   
       call compute_emission_bias(thisOctal,subcell, grid, lambda, lambda_array, n_lambda)
    end do

   
    do ichild = 1, thisoctal%nchildren, 1
      child => thisoctal%child(ichild)
      call assign_emission_bias(child, grid, lambda, lambda_array, n_lambda)
    end do


  end subroutine assign_emission_bias




  !
  ! Recursively all cells which are outside of a cydinder with radius Rmax and its center 
  ! passing through a source (star) with index i and an observer.  
  !
  ! The sources outside of the cylinder is also removed by setting their luminosity to zero.
  !
  ! NB: If the index is zero, it does not do anyting (just exits the program) with the sources
  !     and the grid/octal untouched. 
  recursive subroutine restrict(in_octal, i, nsource, sources, n_obs, Rmax)
    implicit none
    
    type(octal), pointer :: in_octal                    ! this root node of the octal in grid
    integer, intent(in)  :: i                           ! index in the source arrays (stars) 
    integer, intent(in)  :: nsource                     ! The number of sources.
    type(sourcetype), intent(inout) :: sources(nsource) ! sources (stars)
   type(vector), intent(in)    :: n_obs            ! direcion to the observer (should be normalized)
    real(double), intent(in)    :: Rmax        ! The radius of the cylinder (in 10^10cm)
    !
    logical, save :: first_time=.true.
    integer :: j, subcell

    type(vector) :: r0, r, rp
    real(double) :: p
    type(octal), pointer  :: child 

    !
    if (i==0)  then 
       return
    end if

    if (first_time) then
       write(*,*) " "
       write(*,*) "Restrcting the SED calculations to ", i, "-th star in cluster"
       write(*,*) " "
       ! assign the luminosity of the sources to be zero
       ! except for the i-th star.
       do j = 1, nsource
          if (j /= i ) sources(j)%luminosity = 1.0d-30
       end do
       !
       ! To avoid a problem in randomsource routine ..
       if (i==1) then
          ! move the star to the second position...
          sources(2) = sources(1)
          sources(1)%luminosity = 0.0d0
       end if

       first_time=.false.       
    end if

    !    
    do subcell = 1, 8
       if (in_octal%hasChild(subcell)) then
          ! find the child
          do j = 1, in_octal%nChildren
             if (in_octal%indexChild(j) == subcell) then
                child => in_octal%child(j)
                call restrict(child, i, nsource, sources, n_obs, Rmax)
                exit
             end if
          end do
       else
          ! this must be a leaf node, so check if it is 
          ! in the cylinder. 
          ! If so,  set the opacity 
          ! and the emissivity to zero. Set the inflow flag to 
          ! .false. as well. 
          r0 = sources(i)%position
          r = subcellCentre(in_octal,subcell)
          rp = r-r0

          p = ABS(rp .dot. n_obs) 

          if ( p > Rmax) then 
! UNCOMMENT THE NEXT TWO LINES FOR UNREDDEN SIMULATION
!             in_octal%inFlow(subcell) = .false.  
!             in_octal%rho(subcell) = 1.e-30
!
             in_octal%temperature(subcell) = 3.0
             in_octal%velocity = VECTOR(0.,0.,0.)
             in_octal%biasCont3D(subcell) = 1.0
             in_octal%biasLine3D(subcell) = 1.0
             in_octal%etaLine(subcell) = 1.e-30
             in_octal%etaCont(subcell) = 1.e-30
          else
             continue
          end if

!          !  For special case...
!          ! Turning off the dust emission beyond Rmax
!          if (modulus(rp) > Rmax) then
!             in_octal%temperature(subcell) = 3.0
!             in_octal%velocity = VECTOR(0.,0.,0.)
!             in_octal%biasCont3D(subcell) = 1.0
!             in_octal%biasLine3D(subcell) = 1.0
!             in_octal%etaLine(subcell) = 1.e-30
!             in_octal%etaCont(subcell) = 1.e-30
!          end if

       end if 

    end do 

  end subroutine restrict




  !
  ! Recursively reassign the artificial 10 K temperature assigned in lucy's routine
  recursive subroutine reassign_10K_temperature(in_octal)
    implicit none
    
    type(octal), pointer :: in_octal                    ! this root node of the octal in grid
    !
    integer :: j, subcell

    type(octal), pointer  :: child 

    !    
    do subcell = 1, 8
       if (in_octal%hasChild(subcell)) then
          ! find the child
          do j = 1, in_octal%nChildren
             if (in_octal%indexChild(j) == subcell) then
                child => in_octal%child(j)
                call reassign_10K_temperature(child)
                exit
             end if
          end do
       else
          ! this must be a leaf node,
          ! Correcting 10 K dust temperaure. (This is set in
          ! Lucy's routine, but it is contributing to the disc emission

          if (in_octal%temperature(subcell) == 10.0) in_octal%temperature(subcell)=3.0
          
       end if 

    end do

  end subroutine reassign_10K_temperature

end module cluster_class

#endif
